#include <cstdio>
#include <algorithm>
#include <climits>
#include <cfloat>
#include <cstdlib>
#include <vector>
#include <OsiClpSolverInterface.hpp>
//#include <coin/OsiCpxSolverInterface.hpp>
#include <CglClique.hpp>
#include <CglEClique.hpp>
#include <CglOddHole.hpp>
#include <CglGomory.hpp>
#include <CglTwomir.hpp>
#include <stdio.h>
extern "C"
{
#include "strUtils.h"
#include "cgraph.h"
#include "memory.h"
#include "oddhs.h"
#include "clique_separation.h"
}
#include "osi_cgraph.h"
#include "strUtils.h"

using namespace std;

#define MIN_VIOLATION 0.02
#define MAX_TIME 500

/* minimum fractional part to a variable to be considered fractional */
#define MIN_FRAC      0.001

double fracPart( const double x );

/* CBC stuff */

vector< int > fvi; /* fractional variables of interest */

void readLP( const char *fileName );

void printGraphSummary( CGraph *cgraph, const char *graphName );

/* parameters */

typedef enum
{
   Default,   /* our new clique separation implementation */
   CglSepM
} SeparationMethod;

SeparationMethod sepMethod = Default;
char methodName[256];
int maxTwoMir = 0;
int maxGomory = 0;

void help()
{
   printf("usage:\n");
   printf("cbcroot problem [-cgl]\n");
   printf("\t-cgl changes separation routine to CGL\n");
   printf("\t-mtm=int max two mir passes for when no clique cuts are found\n");
   printf("\t-mgm=int max gomory passes for when no clique cuts are found\n");
   clq_sep_params_help_cmd_line();
}

void parseParameters( int argc, char **argv );

static OsiSolverInterface *solver = NULL;

int addOddHoles( OsiSolverInterface *solver, OddHoleSep *oddhs );

OsiClpSolverInterface *realSolver = new OsiClpSolverInterface();

void decideLpMethod();

int main( int argc, char **argv )
{
   if ( argc < 2 )
   {
      help();
      exit( EXIT_FAILURE );
   }

   realSolver = new OsiClpSolverInterface();

   solver = (OsiSolverInterface*) realSolver;
   parseParameters( argc, argv );

   readLP( argv[1] );

   //decideLpMethod();

   printf("cbcroot: root node relaxation and clique cuts\n\n");
   char problemName[ 256 ];
   getFileName( problemName, argv[1] );
   printf("loaded %s \n", problemName );
   printf("\t%d variables (%d integer) %d rows\n\n", solver->getNumCols(), solver->getNumIntegers(), solver->getNumRows() );

   CGraph *cgraph = osi_build_cgraph( solver );

   fvi.resize( solver->getNumCols() );

   {
      CliqueSeparation *clqSep = clq_sep_create( cgraph );
      //clq_sep_set_verbose( clqSep, 1 );
      clq_sep_set_params_parse_cmd_line( clqSep, argc, (const char**)argv );
      clq_sep_params_print( clqSep );
      clq_sep_free( &clqSep );
   }
   CliqueSeparation *clqSep = NULL;


   int pass = 0;
   int newCuts = 0;
   vector<double> ones( solver->getNumCols(), 1.0 );
   const CliqueSet *clqSet = NULL;

   printf("solving relaxation ... ");
   clock_t start = clock();
   //solver->setHintParam( OsiDoDualInInitial, true, OsiHintDo );
   solver->initialSolve();

   if (!solver->isProvenOptimal())
   {
      fprintf( stderr, "ERROR: Could not solve LP relaxation. Exiting.\n" );
      exit( EXIT_FAILURE );
   }

   printf("Initial dual bound %g\n", solver->getObjValue() );
   clock_t end = clock();
   printf("solved in %.3f\n", ((double)(end-start)) / ((double)CLOCKS_PER_SEC) );
   //clock_t startSep = 0, endSep;
   double timeSep;

   CliqueSeparation *cs = clq_sep_create( cgraph );
   clq_sep_set_params_parse_cmd_line( cs, argc, (const char**)argv );
   const int maxPasses = clq_sep_get_max_passes(cs);
   clq_sep_free( &cs );

   int passesGomory = 0;
   int passesTwoMir = 0;

   do
   {
      newCuts = 0;

      //const double *x = solver->getColSolution();

      clock_t startSep = clock();

      switch (sepMethod)
      {
      case Default :
         /* updating reduced costs */
      {
         CglEClique cliqueGen;
         OsiCuts cuts;
         CglTreeInfo info;
         info.level = 0;
         info.pass = 1;

         cliqueGen.parseParameters( argc, (const char**)argv );
         cliqueGen.setCGraph( cgraph );
         cliqueGen.setGenOddHoles( false );
         cliqueGen.colNames = &(solver->getColNames());

         cliqueGen.generateCuts( *solver, cuts, info );
         newCuts = cuts.sizeCuts();

         solver->applyCuts( cuts );
      }


      break;
      case CglSepM :
      {
         CglClique cliqueGen;
         OsiCuts cuts;
         CglTreeInfo info;
         info.level = 0;
         info.pass = 1;
         cliqueGen.setMinViolation( MIN_VIOLATION );
         
         cliqueGen.generateCuts( *solver, cuts, info );
         newCuts = cuts.sizeCuts();

         solver->applyCuts( cuts );

      }

      break;
      }

      clock_t pend = clock();
      double pTime = ((double)(pend-start)) / ((double)CLOCKS_PER_SEC);
      double sepTime = ((double)(pend-startSep)) / ((double)CLOCKS_PER_SEC);


      if ( pTime > MAX_TIME )
      {
         printf("time limit reached. discarding cuts.\n");
         newCuts = 0;
      }
      printf( "round %d : %s separated %d cuts in %.4f seconds. dual limit now: %g time: %.3f\n", pass+1, methodName, newCuts, sepTime, solver->getObjValue(), pTime );

      if (newCuts<60)
      {
         if (passesGomory<maxGomory)
         {
            printf("\n- generating gomory cuts.\n");
            CglGomory cglGomory;
            OsiCuts cuts;
            CglTreeInfo info;
            info.level = 0;
            info.pass = 1;
            //            cglGomory.setMinViolation( MIN_VIOLATION );

            cglGomory.generateCuts( *solver, cuts, info );
            newCuts = cuts.sizeCuts();

            newCuts += cuts.sizeCuts();

            solver->applyCuts( cuts );
            ++passesGomory;
         }
         if (passesTwoMir<maxTwoMir)
         {
            printf("\n- generating two mir cuts.\n");
            CglTwomir cglTwoMir;
            OsiCuts cuts;
            CglTreeInfo info;
            info.level = 0;
            info.pass = 1;
            //            cglTwoMir.setMinViolation( MIN_VIOLATION );

            cglTwoMir.generateCuts( *solver, cuts, info );
            newCuts = cuts.sizeCuts();

            solver->applyCuts( cuts );
            newCuts += cuts.sizeCuts();

            ++passesTwoMir;
         }
      }

      ++pass;

      if (newCuts)
      {
         printf("resolving relaxation ... ");
         fflush( stdout );
         clock_t start = clock();
         solver->resolve();
         if (!solver->isProvenOptimal())
         {
            fprintf( stderr, "ERROR: Could not solve LP relaxation. Exiting.\n" );
            exit( EXIT_FAILURE );
         }
         clock_t end = clock();
         printf("solved in %.3f after adding cuts of pass %d. dual limit now %g\n", ((double)(end-start)) / ((double)CLOCKS_PER_SEC), pass, solver->getObjValue() );
         printf("\n");
         fflush( stdout );
      }

      fflush( stdout );

   }
   while ( (newCuts>0) && (pass<maxPasses) ) ;

   clock_t tend = clock();
   double totalTime = ((double)(tend-start)) / ((double)CLOCKS_PER_SEC);
   printf("\nEnd of root node relaxation. Dual limit: %g time: %.3f \n", solver->getObjValue(), totalTime );

   //printf("min violation: %g\n", cliqueGen.getMinViolation() );

   /*clq_sep_free( &clqSep );*/
   cgraph_free( &cgraph );

   delete realSolver;

   return EXIT_SUCCESS;
}

void readLP( const char *fileName )
{
   solver->setIntParam(OsiNameDiscipline, 2);
   solver->readMps( fileName );
   solver->setIntParam(OsiNameDiscipline, 2);
   solver->messageHandler()->setLogLevel(1);
   solver->setHintParam(OsiDoReducePrint,true,OsiHintTry);
}

void printGraphSummary( CGraph *cgraph, const char *graphName )
{
   vector< int > neighs;
   neighs.reserve( cgraph_size(cgraph)*100 );
   printf("\n");
   printf("Summary for graph: %s\n", graphName );

   int minD = INT_MAX;
   int maxD = -1;
   double minF = DBL_MAX;
   double maxF = -1;
   double mostF = 0.0;

   int nEdges = 0;
   for ( int i=0 ; (i<cgraph_size(cgraph)) ; ++i )
   {
      int nConf = cgraph_get_all_conflicting( cgraph, i, &(neighs[0]), solver->getNumCols()*100 );
      nEdges += nConf;

      if ( nConf > maxD )
         maxD = nConf;
      if ( nConf < minD )
         minD = nConf;

      double f = ((double)cgraph_get_node_weight(cgraph,i)) / 1000.0;

      if ( f < minF )
         minF = f;
      if ( f > maxF )
         maxF = f;

      //const double *lb = solver->getColLower();
      //const double *ub = solver->getColUpper();
      if ( solver->isBinary(cgraph_get_original_node_index(cgraph, i)) )
      {
         double fracPart = std::min( 1.0-f, f );
         if ( fracPart > mostF )
            mostF = fracPart;
      }

   }
   nEdges /= 2;

   printf("\tnodes : %d\n", cgraph_size(cgraph) );
   printf("\tedges : %d\n", nEdges );
   printf("\tminimum degree: %d\n", minD);
   printf("\tmaximum degree: %d\n", maxD);
   printf("\tminF: %.4f\n", minF);
   printf("\tmaxF: %.4f\n", maxF);
   printf("\tmostF: %.4f\n", mostF);
   printf("\n");
}

double fracPart( const double x )
{
   double nextInteger = ceil( x );
   double previousInteger = floor( x );

   return std::min( nextInteger-x, x-previousInteger );
}

void parseParameters( int argc, char **argv )
{
   int i;
   sprintf( methodName, "def" );
   for ( i=1 ; (i<argc) ; ++i )
   {
      if (strstr( argv[i], "-cgl"))
      {
         printf("Using COIN CGL separation.\n");
         sprintf( methodName, "cgl" );
         sepMethod = CglSepM;
         continue;
      }

      char pName[256];
      char pValue[256];
      getParamName( pName, argv[i] );
      getParamValue( pValue, argv[i] );

      if (strcmp( pName, "mtm" )==0)
      {
         maxTwoMir = atoi( pValue );
         printf("using at max %d two mir passes.\n", maxTwoMir );
         continue;
      }

      if (strcmp( pName, "mgm" )==0)
      {
         maxGomory = atoi( pValue );
         printf("using at max %d gomory passes.\n", maxGomory );
         continue;
      }

   }
}

int addOddHoles( OsiSolverInterface *solver, OddHoleSep *oddhs )
{
   int r = 0;
   vector< double > row( solver->getNumCols(), 1.0);
   OsiCuts oc;
   for ( int i=0; (i<oddhs_get_odd_hole_count(oddhs)) ; i++ )
   {
      const int *s = oddhs_get_odd_hole( oddhs, i );
      const int *e = oddhs_get_odd_hole( oddhs, i+1 );
      const int size = e-s;

      double lhs = 0.0;
      for ( int j=0; (j<size) ; j++ )
         lhs +=  solver->getColSolution()[s[j]];

      const double viol = lhs - ((double)(size/2));
      if (viol < 1e-5)
         continue;

      printf("lhs %g viol %g rhs %g\n", lhs, viol, (double)(size/2) );

      OsiRowCut orc;
      orc.setRow( size, s, &row[0] );
      orc.setUb( ((double)(size/2)) );
      orc.setLb( -COIN_DBL_MIN );
      orc.setGloballyValid( true );

      oc.insertIfNotDuplicate( orc );
      r++;
   }

   solver->applyCuts( oc );

   return r;
}

void decideLpMethod()
{
   if ( realSolver->getNumRows() > 100000 )
   {
      ClpSolve::SolveType method = ClpSolve::useBarrier;
      ClpSolve::PresolveType presolveType = ClpSolve::presolveOn;
      int numberPasses = 5;
#ifndef UFL_BARRIER
      int options[] = {0,0,0,0,0,0};
#else
      // we can use UFL code
      int options[] = {0,0,0,0,4,0};
#endif
      int extraInfo[] = {-1,-1,-1,-1,-1,-1};
      int independentOptions[] = {0,0,3};
      ClpSolve clpSolve(method,presolveType,numberPasses,
                        options,extraInfo,independentOptions);
      // =======================
      // now pass options in
      realSolver->setSolveOptions(clpSolve);
   }
}

